#!/usr/bin/env python
"""
_LoadBlocks_

Oracle implementation of DBS3Buffer.LoadBlocks
"""

from WMComponent.DBS3Buffer.MySQL.LoadBlocks import LoadBlocks as MySQLLoadBlocks

class LoadBlocks(MySQLLoadBlocks):
    pass
