#!/usr/bin/env python
"""
_CountPhedexNotUploaded_

Oracle implementation of DBSBuffer.CountPhedexNotUploaded
"""

from __future__ import division
from WMComponent.DBS3Buffer.MySQL.CountPhedexNotUploaded import CountPhedexNotUploaded as MySQLCountPhedexNotUploaded

class CountPhedexNotUploaded(MySQLCountPhedexNotUploaded):
    """
    CountPhedexNotUploaded

    """
    pass
