#!/usr/bin/env python
"""
_SetBlockFiles_

Oracle implementation of DBS3Buffer.SetBlockFiles
"""

from WMComponent.DBS3Buffer.MySQL.SetBlockFiles import SetBlockFiles as MySQLSetBlockFiles

class SetBlockFiles(MySQLSetBlockFiles):
    pass
