#!/usr/bin/env python
"""
_Dummy_

A dummy class to mimic the component for tets.
"""




import logging

class Dummy:
    """
    _Dummy_

    A dummy class to mimic the component for tets.
    """

    def __init__(self):
        logging.debug("I am a dummy object!")
        self.config = None
