#!/usr/bin/env python
"""
brokenCmsRun.py

cmsRun simulator that exits with a non zero exit code.

Created on Feb 26, 2013

@author: dballest
"""

import sys

if __name__ == '__main__':
    sys.exit(65)
