#!/usr/bin/env python
"""
_GetFinishedSubscriptions_

Oracle implementation of Subscriptions.GetFinishedSubscriptions

Created on Aug 29, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Subscriptions.GetFinishedSubscriptions import GetFinishedSubscriptions as MySQLFinishedSubscriptions

class GetFinishedSubscriptions(MySQLFinishedSubscriptions):
    pass
