#!/usr/bin/env python
"""
_CountOpenBlocks_

Oracle implementation of DBSBuffer.CountOpenBlocks
"""

from __future__ import division
from WMComponent.DBS3Buffer.MySQL.CountOpenBlocks import CountOpenBlocks as MySQLCountOpenBlocks

class CountOpenBlocks(MySQLCountOpenBlocks):
    """
    _CountOpenBlocks_

    """
    pass
