"""
_CountFinishedSubscriptionsByTask_

Oracle implementation of CountFinishedSubscriptionsByTask
"""
from WMCore.WMBS.MySQL.Subscriptions.CountFinishedSubscriptionsByTask \
     import CountFinishedSubscriptionsByTask as MySQLCountFinishedSubscriptionsByTask

class CountFinishedSubscriptionsByTask(MySQLCountFinishedSubscriptionsByTask):
    pass
