#!/usr/bin/env python
"""
_UpdateAlgoDatasetAssoc_

Oracle implementation of DBSBuffer.UpdateAlgoDatasetAssoc
"""




from WMComponent.DBS3Buffer.MySQL.UpdateAlgoDatasetAssoc import UpdateAlgoDatasetAssoc as MySQLUpdateAlgoDatasetAssoc

class UpdateAlgoDatasetAssoc(MySQLUpdateAlgoDatasetAssoc):
    """
    _UpdateAlgoDatasetAssoc_

    Update the in_dbs column for a particular algo/dataset association.
    """
    pass
